' Minimal SD booter Test
' ----------------------
' (c)2012 Andy Schenk, Insonix
' RR20120820 modified by Cluso99 to test SD cards (for later use in P2 ROM)
' RR20120824 display a few more details on MBR
' RR20160901 try onP8XBlade2
'-----------------------------------------------
' rename SD_test_100.spin       P8XBlade2 working w both SanDisk Ultra HCI Class 10 8GB & Mixza 8GB HCI(1)
' RR20160927 v100
'            v101a  add SD_trace_032d & fix buffers at $4000 hub
'                c  wkg w/o SD_trace_032d
' v101d_1a          log more data
'        c          why does /CS=1 before CMD41 ???
'        d          stop CMD41 from /CS=1
'        h          also counts & times cmd55+cmd41 sent 
'        j          + sector #1 display - wkg :)
' RR20160929           not wkg so go back to SD_test_101d_1k and copy parts from SD_test_032e-1l
'        l,m,p      wkg
' v101d_1q10        both cards wkg
'       1q20        wkg for SDV2(block address); SDV2(byte address not verified); SDV1(not coded correctly); MMC(ignored)
'       1q22  (bug) add $FF stall counter; SDV2(block & byte) & SDV1 should work correctly; MMC ignored/error
'         21->25    add $FF stall counter & $77777777 identifier
'         29        wkg
' RR20161003 v102   wkg (need to test SDV2(byte address) & SDV1 cards)
'                d  fix: SanDisk 2GB rejects CMD16
'                f  fix: remove shl blockno,blocksh from cmd to readblock routine (caused reject of CMD16)
'                   fix: CMD16 does not change blocksh to 0!!
' RR20161003     g  wkg SDV2(block & byte) (1GB & 2GB SDV2byte, 8GB & 64GB SDV2block) check ok
' RR20161003 v103   P1 test release
' RR20161012 v104   CMD8 $05 reply reqd to test for SDV1
'                   dane-elec sdhc card did not initialise (fails CMD0) - add try 74clocks with CS=0 & 50 retries
'                   74clocks CS=0+DI=1 fails on 1GB SD, try 96 x1's OK
'                   P1 test release


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  _freq    = 80_000_000

   do  = 0      'sd pins
   clk = 1
   di  = 2
   cs  = 3

{{
'' P8XBlade2
  _clkmode = xtal1 + pll8x
  _xinfreq = 12_000_000
  _freq    = 96_000_000

   do  = 12      'sd pins
   clk = 13
   di  = 14
   cs  = 15
}}

{{
'RamBlade3
  _clkmode  = xtal1 + pll16x
  _xinfreq  = 6_500_000
  _freq     = 104_000_000

  DO     = 20                'microSD
  Clk    = 21
  DI     = 22
  CS     = 23
}}


' Command             Argument        Response/Data    Description
  CMD0  = 0  +$40   ' 0               R1       -       GO_IDLE_STATE            *Reqs valid CRC      
  CMD1  = 1  +$40   ' 0               R1       -       SEND_OP_COND
  ACMD41= 41 +$40   ' $4000_0000      R1       -       APP_SEND_OP_COND         *Reqs CMD55 first
  CMD8  = 8  +$40   ' 0               R1+R7    -       SEND_IF_COND             *Reqs valid CRC
  CMD9  = 9  +$40   ' 0               R1       Y       SEND_CSD
  CMD10 = 10 +$40   ' 0               R1       Y       SEND_CID
  CMD12 = 12 +$40   ' 0               R2       -       STOP_TRANSMISSION
  CMD16 = 16 +$40   ' BlkLen[31:0]    R1       -       SET_BLOCKLEN
  CMD17 = 17 +$40   ' Addr[31:0]      R1       Y       READ_SINGLE_BLOCK
  CMD18 = 18 +$40   ' Addr[31:0]      R1       Y       READ_MULTIPLE_BLOCK
  CMD23 = 23 +$40   ' NoBlks[15:0]    R1       -       SET_BLOCK_COUNT
  ACMD23= 23 +$40   ' NoBlks[22:0]    R1       -       SET_WR_BLOCK_ERASE_COUNT *Reqs CMD55 first
  CMD24 = 24 +$40   ' Addr[31:0]      R1       Y       WRITE_BLOCK
  CMD25 = 25 +$40   ' Addr[31:0]      R1       Y       WRITE_MULTIPLE_BLOCK
  CMD55 = 55 +$40   ' 0               R1       -       APP_CMD                  *Prefix for ACMD41/ACMD23
  CMD58 = 58 +$40   ' 0               R1+R3    -       READ_OCR  

'Hub Trace Addresses
  done          = $4000         'pasm sets done[0] to $00 when done
  mbr           = done + 4      'pasm stores the MBR sector here
  vol           = mbr  + 512    'pasm stores the VOL sector here
  param         = vol  + 512    'pasm stores (cmd + par + reply + $FF) for each cmd


OBJ
        dbg:    "FullDuplexSerial"

VAR
  long  SD_type
  
pub Main : i | tm
'
' This Spin code is not part of the boot code, it
' just starts the PASM cog and shows the result
'
  waitcnt(clkfreq*3+cnt)
  dbg.start(31,30,0,115200)
  waitcnt(cnt+clkfreq/1000)
  dbg.str(string(13,"SD testing v104",13))

  longfill(done,$9999_9999,$3C00/4)                     ' clear the buffers from hub $4000
  byte[done] := $FF                                     ' set busy
                           
  cognew(@entry, done)                                  ' start the SD Driver

  repeat until (byte[done][0] == 0)

  DisplayParam
  DisplayMBR
' DisplayVOL

  dbg.str(string(13,"SD card: "))
  case SD_type
    1 : dbg.str(string("1 = V1",13))
    2 : dbg.str(string("2 = V2(byte address)",13))
    3 : dbg.str(string("3 = SDHC/SD V2(block address)",13))
    other : dbg.str(string("Unknown ???",13))
  
  repeat                                                '<========


Pri DisplayParam | i, j, x, y
  dbg.str(string("Parameters used...",13))

  repeat i from 0 to $3C00 '$800*4 '7800/4 '7000/4 '$3C00/4 '255
    x := long[param][i]
    case x
      $FFFF_FFFF: 
          dbg.str(string(";",13))
          waitcnt(cnt+clkfreq/100)
          dbg.hex(i*4+param,4)
          dbg.str(string(": "))
      $FF11_11FF: 
          dbg.str(string("nCS=1    ;",13))
          waitcnt(cnt+clkfreq/100)
          dbg.hex(i*4+param,4)
          dbg.str(string(": "))
      $7777_7777:                                       ' counter follows...
          y := long[param][++i]
          dbg.tx("(")
          dbg.dec(y)
          dbg.str(string(") "))
      $8888_8888:                                       ' CMD17/18/24 $FE Data Token
          dbg.str(string(";",13))
          waitcnt(cnt+clkfreq/100)
          dbg.hex(i*4+param,4)
          dbg.str(string(": CRC..... "))
      $9999_9999: 
          dbg.str(string("<eof>",13))
          SD_type := long[param][++i]
          waitcnt(cnt+clkfreq/100)
          return 0
      other:
          dbg.hex(x,8)
          dbg.tx(" ")


Pri DisplayMBR | r, i, j
  dbg.str(string(13,"MBR sector 0...",13))
' repeat j from 0 to 31
  repeat j from 28 to 31        '<========================
    DisplayHex(mbr, j*16)

{{
'display the 16 bytes we might use immediately preceeding the ptn tables
  dbg.str(string("$1AE(16): "))
  repeat i from 0 to 15
    dbg.hex(byte[mbr][$1AE + i],2)
  dbg.tx(13)
}}

'display the ptn tables if it is an MBR
  if byte[mbr][510] == $55 and byte[mbr][511] == $AA
    dbg.str(string("Valid MBR",13))
    repeat j from 0 to 3
      dbg.str(string("PTN "))
      dbg.dec(j)
      dbg.str(string(": Typ=$"))
      dbg.hex(byte[mbr][$1BE+$04+ j*16],2)
      dbg.str(string("  "))
'     dbg.tx(13)
      DisplayHex(mbr, $1BE + j*16)
  else
    dbg.str(string("Not MBR",13))

Pri DisplayVOL | r, i, j
  dbg.str(string("VOL sector 1...",13))
  repeat j from 0 to 31
    DisplayHex(vol, j*16)

PRI DisplayHex(adr, j) | i
    dbg.hex(j,4)                                        ' address within sector
    dbg.str(string(": "))
    repeat i from 0 to 15                               ' hex
      dbg.hex(byte[adr][i + j],2)
      dbg.tx(" ")
    dbg.tx(" ")
    repeat i from 0 to 15                               ' ascii
      if byte[adr][i + j] < $20 or byte[adr][i + j] > $7E
        dbg.tx(".")
      else
        dbg.tx(byte[adr][i + j])
    dbg.tx(13)


DAT
                ORG     000
entry
                mov     ptrdone,          par           ' @done
                mov     ptrmbr,           ptrdone
                add     ptrmbr,           #4            ' @mbr
                mov     ptrvol,           ptrmbr
                add     ptrvol,           sectorsz      ' @vol
                mov     ptrcmd,           ptrvol
                add     ptrcmd,           sectorsz      ' @param

        
'+-----------------------------------------------------------------------------+
'+      SD/SDHC SPI Initialisation                                             +
'+-----------------------------------------------------------------------------+

'+-----------------------------------------------------------------------------+
'+      Wait 1ms (we can assume this has been done)                            +
'+-----------------------------------------------------------------------------+
                mov     ctr1,             cnt
                add     ctr1,             delay1ms
                waitcnt ctr1,             zero

'+-----------------------------------------------------------------------------+
                wrlong  xFFs,             ptrcmd        '\\                     FFFFFFFF
                add     ptrcmd,           #4            '//
'+-----------------------------------------------------------------------------+

'+-----------------------------------------------------------------------------+
'+      Send >74clocks with /CS=1 & DI=1                                       +
'+-----------------------------------------------------------------------------+
                mov     outa,             initmask      ' 1<<cs | 1<<di | 0<<clk 
                mov     dira,             dirmask       ' 1<<cs | 1<<di | 1<<clk (DO=input to prop)
' 74 clocks with /CS=1 & DI=1 starting with CLK=0
                mov     ctr1,             #74
                mov     ctr2,             cnt           '\ delay 10us = 100KHz/2
                add     ctr2,             delay10us     '/
:count74        andn    outa,             clkmask       ' CLK=0
                waitcnt ctr2,             delay10us     
                or      outa,             clkmask       ' CLK=1
                waitcnt ctr2,             delay10us
                djnz    ctr1,             #:count74
'+-----------------------------------------------------------------------------+
                andn    outa,             clkmask       ' CLK0 (idle)  & /CS=1
'+-----------------------------------------------------------------------------+
{{
'+-----------------------------------------------------------------------------+
'+      Send >74clocks with /CS=0 & DI=1                                       +
'+-----------------------------------------------------------------------------+
'''             mov     outa,             initmask      ' 1<<cs | 1<<di | 0<<clk 
'''             mov     dira,             dirmask       ' 1<<cs | 1<<di | 1<<clk (DO=input to prop)
' 74 clocks with /CS=0 & DI=1 starting with CLK=0
                andn    outa,             csmask        ' /CS=0 (enable)
                mov     ctr1,             #74
                mov     ctr2,             cnt           '\ delay 10us = 100KHz/2
                add     ctr2,             delay10us     '/
:count74a       andn    outa,             clkmask       ' CLK=0
                waitcnt ctr2,             delay10us     
                or      outa,             clkmask       ' CLK=1
                waitcnt ctr2,             delay10us
                djnz    ctr1,             #:count74a
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                andn    outa,             clkmask       ' CLK0 (idle)  & /CS=1
'+-----------------------------------------------------------------------------+
}}

' Send 96 clocks with /CS=0 & DI=1
'+-----------------------------------------------------------------------------+
                andn    outa,             csmask        ' /CS=0 (enable)
'+-----------------------------------------------------------------------------+
                neg     dataout,          #1            '\ presend 3*32x 1's 
                mov     bitsout,          #96           '|
                call    #sendio                         '/
'+-----------------------------------------------------------------------------+
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                mov     ctr2,             cnt           '\ delay 10us = 100KHz/2
                add     ctr2,             delay10us     '/
                waitcnt ctr2,             delay10us
'+-----------------------------------------------------------------------------+



                
'+-----------------------------------------------------------------------------+
'+ Software Reset:                                                             +
'+      CMD0,  PAR=$0,        CRC=$95, REPLY=$01                               +
'+-----------------------------------------------------------------------------+
Command0        mov     blocksh,          #0            ' block shift =0
                mov     starttime,        cnt
                mov     ctr1,             #50           '#9 ' try a few times
:again0         mov     cmdout,           #CMD0
                mov     cmdpar,           #0
                mov     cmdcrc,           #$95
                call    #cmd
'+-----------------------------------------------------------------------------+
                wrlong  xFFs, ptrcmd                    '\\                     FFFFFFFF
                add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
                cmp     reply,            #1    wz      ' reply = $01 ?
        if_z    jmp     #Command8                       ' y: SD/MMC card, but not fully validated
'+-----------------------------------------------------------------------------+
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                djnz    ctr1,             #:again0      ' n: try again?
                jmp     #fail00                         ' failure: failed CMD0
'+-----------------------------------------------------------------------------+
' we know we now have an SD/MMC card but its not fully validated yet

'+-----------------------------------------------------------------------------+
'+ Check Voltage:                                                              +
'+      CMD8,  PAR=$1AA,      CRC=$87, REPLY=R1($01)+R7($xx1AA) ($05=try SDV1) +
'+-----------------------------------------------------------------------------+
Command8        mov     cmdout,           #CMD8         ' 
                mov     cmdpar,           x000001AA
                mov     cmdcrc,           #$87
                call    #cmd
'+-----------------------------------------------------------------------------+
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                cmp     reply8,           #$05      wz  ' $05 ?
        if_z    jmp     #:SDV1                          ' y: cmd not recognised, try SDV1
                cmp     reply8,           #1        wz  ' $01 ?
        if_nz   jmp     #fail01                         ' n: failure: unknown reply
                and     reply,            xFFF          ' lower 12bits
                cmp     reply,            x000001AA wz  ' same?
        if_z    mov     cmdpar2,          x40000000     ' y: SDV2
        if_z    jmp     #Command55                      ' y: CMD55+ACMD41(bit30)
                jmp     #fail02                         ' n: unknown R7
                
:SDV1           mov     cmdpar2,          #0            ' n: SDV1
                mov     blocksh,          #9            ' n: byte addressing
                jmp     #Command55                      ' n: CMD55+ACMD41(0)
'+-----------------------------------------------------------------------------+

'+-----------------------------------------------------------------------------+
'+ Check SD-V2:                                                                +
'+      CMD55, PAR=$0,        CRC=$xx, REPLY=$01                               +
'+     ACMD41, PAR=$40000000, CRC=$xx, REPLY=$00 SD-V2                         +
'+     ACMD41, PAR=$0,        CRC=$xx, REPLY=$00 SD-V1                         +
'+-----------------------------------------------------------------------------+
Command55       mov     ctr5541,          #0            '\\                     count # times CMD55+41 sent
                mov     saveptrcmd,       ptrcmd        '||                     current pointer
                mov     time5541,         cnt           '//
:again55        mov     cmdout,           #CMD55
                mov     cmdpar,           #0
                call    #cmd
                cmp     reply,            #1    wz
        if_nz   jmp     #fail05
'+-----------------------------------------------------------------------------+
'''             wrlong  xFFs, ptrcmd                    '\\                     FFFFFFFF
'''             add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
                mov     cmdout,           #ACMD41
                mov     cmdpar,           cmdpar2       ' SDV2=$40000000 / SDV1=0
                call    #cmd
'-------------------------------------------------------------------------------
                or      outa,             csmask        ' /CS=1 (disable)
'-------------------------------------------------------------------------------
                add     ctr5541,          #1            ' y: cnt5541++
                wrlong  x77s,             ptrcmd        '\\                     count follows... (CMD55+41 sent #)
                add     ptrcmd,           #4            '//
                wrlong  ctr5541,          ptrcmd        '\\                     no times CMD55+41 sent
                add     ptrcmd,           #4            '//
                mov     timer,            cnt
                sub     timer,            time5541
                wrlong  x77s,             ptrcmd        '\\                     count follows... (time in clocks)
                add     ptrcmd,           #4            '//
                wrlong  timer,            ptrcmd        '\\                     no. of clocks while CMD55+41 sent
                add     ptrcmd,           #4            '//
                wrlong  xFFs,             ptrcmd        '\\                     FFFFFFFF
                add     ptrcmd,           #4            '//
'-------------------------------------------------------------------------------
                cmp     reply,            #1    wz      
        if_z    mov     ptrcmd,           saveptrcmd    ' y: busy - reposition to start of CMD55+CMD41 sequence
        if_z    jmp     #:again55                       ' y: busy
                cmp     reply,            #0    wz
        if_nz   jmp     #fail04                         ' n: failure - unknown card

                cmp     cmdpar2,          #0    wz      ' SDV1 ?
        if_z    mov     SDtype,           #1            ' y: SDV1
        if_z    jmp     #Command16                      ' y: SDV1 does not use CMD58
                                                        ' n: SDV2 (fall thru')

'+-----------------------------------------------------------------------------+
'+ Check OCR CCS bit:                                                          +
'+      CMD58, PAR=$0,        CRC=$xx, REPLY=R1($00)+R3(b30=1)                 +
'+-----------------------------------------------------------------------------+
Command58       mov     cmdout,           #CMD58        ' SDHC ?
                mov     cmdpar,           #0
                call    #cmd
'+-----------------------------------------------------------------------------+
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                cmp     reply8,           #0    wz      ' $00 ?
        if_nz   jmp     #fail08                         ' n: failure - unknown reply
                test    reply,            x40000000 wz  ' CCS=1?
        if_nz   mov     SDtype,           #3            ' y: SDHC/SDV2(block address)
        if_nz   jmp     #readMBR                        ' y: SDV2/SDHC(blockaddress, blocksize=512bytes) (blocksh=0)
'+-----------------------------------------------------------------------------+
                mov     SDtype,           #2            ' n: SDV2(byte address)
                mov     blocksh,          #9      
                                                        ' n: SDV2(byte address) (fall thru')

'+-----------------------------------------------------------------------------+
'+ Force block size to 512 bytes:                                              +
'+      CMD16, PAR=$200,      CRC=$xx, REPLY=R1($00)                           +
'+-----------------------------------------------------------------------------+
Command16
:again16        mov     cmdout,           #CMD16        ' force blocksize=512bytes
                mov     cmdpar,           sectorsz      ' 512 bytes
                call    #cmd
'+-----------------------------------------------------------------------------+
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                cmp     reply,            #1    wz      
        if_z    jmp     #:again16                       ' n: busy
                cmp     reply,            #0    wz
        if_z    jmp     #readMBR                        ' y: OK
                jmp     #fail06
'+-----------------------------------------------------------------------------+

'+-----------------------------------------------------------------------------+
'+ Read MBR (Sector 0):                                                        +
'+-----------------------------------------------------------------------------+
readMBR         mov     blocknr,          #0            ' MBR = SECTOR 0
                mov     bufad,            ptrmbr
                call    #readblock                      ' read sector 0
{{
readVOL         mov     blocknr,          #1            ' VOL = SECTOR 1
'               mov     blocknr,          PTN0LBA       ' LBA PTN 0 = x00200000
                mov     bufad,            ptrvol
                call    #readblock
}}
                jmp     #alldone
'+-----------------------------------------------------------------------------+


'+-----------------------------------------------------------------------------+
'+ Read Block/Sector:                                                          +
'+      CMD17, PAR=blocknr,   CRC=$xx, REPLY=$FE+block+CRC16                   +
'+-----------------------------------------------------------------------------+
readblock       mov     ctr2,             sectorsz      ' read block
                mov     starttime,        cnt
                mov     cmdout,           #CMD17 
                mov     cmdpar,           blocknr
                shl     cmdpar,           blocksh       ' <<0 or <<9
                call    #cmd
                call    #getreply                       ' n*$FF+$FE
                cmp     reply,            #$FE  wz      ' validate Data Token
        if_nz   jmp     #fail_readblock
'+-----------------------------------------------------------------------------+
                wrlong  x88s, ptrcmd                    '\\                     88888888
                add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
:readbyte       mov     ctr1,             #8            ' 8 bits
                mov     reply,            #0            ' clear reply (not really necessary)
:readbit        andn    outa,             clkmask       ' \ CLK=0
                or      outa,             clkmask       ' / CLK=1
                test    domask,           ina   wc      ' sample on CLK rising edge
                addx    reply,            reply
                djnz    ctr1,             #:readbit     ' bit--
                wrbyte  reply,            bufad         ' save byte
                add     bufad,            #1
                djnz    ctr2,             #:readbyte    ' byte--
                call    #sendiohi                       ' \ read CRC16 1/2
'+-----------------------------------------------------------------------------+
                wrlong  reply, ptrcmd                   '                       Debug: CRC
                add     ptrcmd, #4                      '                       
'+-----------------------------------------------------------------------------+
                call    #sendiohi                       ' / read CRC16 2/2
'+-----------------------------------------------------------------------------+
                wrlong  reply, ptrcmd                   '                       Debug: CRC
                add     ptrcmd, #4                      '                       
'+-----------------------------------------------------------------------------+
                or      outa,             csmask        ' /CS=1 (disable)
'+-----------------------------------------------------------------------------+
                wrlong  xCS1, ptrcmd                    '\\                     FF1111FF
                add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
readblock_ret   ret
'+-----------------------------------------------------------------------------+


'+-----------------------------------------------------------------------------+
'+      SEND: CMDx, PARx, CRCx, GET REPLY                                      +
'+-----------------------------------------------------------------------------+
cmd
'+-----------------------------------------------------------------------------+
                test    csmask, ina       wz            ' /CS=1 ?
        if_nz   wrlong  xCS1, ptrcmd                    '\\                     if /CS=1 then FF1111FF
        if_nz   add     ptrcmd, #4                      '//         
                wrlong  cmdout, ptrcmd                  '\\                     save CMD
                add     ptrcmd, #4                      '//
                wrlong  cmdpar, ptrcmd                  '\\                     save PARAM
                add     ptrcmd, #4                      '//
                wrlong  cmdcrc, ptrcmd                  '\\                     save CRC
                add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
                andn    outa,             csmask        ' /CS=0 (enable)
'+-----------------------------------------------------------------------------+
                neg     dataout,          #1            ' presend 8x 1's  (mandatory!!!)
                call    #sendiohi
'+-----------------------------------------------------------------------------+
                mov     dataout,          cmdout        ' CMD
                call    #sendio                         ' send cmd byte
                mov     dataout,          cmdpar        ' Parameter
                call    #sendiohi                       ' send byte 0
                call    #sendiohi                       ' send byte 1
                call    #sendiohi                       ' send byte 2
                call    #sendiohi                       ' send byte 3
                mov     dataout,          cmdcrc        ' CRC
                call    #sendio                         ' send crc byte
'+-----------------------------------------------------------------------------+

{   ?????????? CMD12 is STOP TRANSMISSION
        cmp cmdout,#CMD12 wz                              '??? CMD12 ???
  if_z  neg dataout,#1
  if_z  call #sendiohi
}

'+-----------------------------------------------------------------------------+
getreply        mov     ctrbusy,          #0            ' reset busy counter
:again          neg     dataout,          #1            ' Receive a Byte Reply (while we output 1's)
                call    #sendiohi
                call    #checktime                      ' timeout?
                cmp     reply,            #$FF   wz     ' if Reply=$FF, continue trying until timeout
        if_z    add     ctrbusy,          #1            ' ctrbusy++
        if_z    jmp     #:again
'+-----------------------------------------------------------------------------+
                cmp     ctrbusy,          #0     wz      ' busy counter >1
        if_z    jmp     #:next
                wrlong  xFF,              ptrcmd        '\\                     save R1 $FF reply
                add     ptrcmd,           #4            '//
                wrlong  x77s,             ptrcmd        '\\                     count follows... ($FF busy)
                add     ptrcmd,           #4            '//
                wrlong  ctrbusy,          ptrcmd        '\\                     busy count
                add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
:next           wrlong  reply,            ptrcmd        '\\                     save R1 reply
                add     ptrcmd,           #4            '//
'+-----------------------------------------------------------------------------+
                cmp     cmdout,           #CMD8  wz
        if_nz   cmp     cmdout,           #CMD58 wz
        if_nz   jmp     #cmd_ret                        ' ret if not CMD8/CMD58
'+-----------------------------------------------------------------------------+
'  CMD8/CMD58: read +R7/R3 reply                                               +
'+-----------------------------------------------------------------------------+
                mov     reply8,           reply         ' save 8bit reply
                mov     bitsout,          #32           
                call    #sendiocnt                      ' read R7/R3 (volts/OCR)
'+-----------------------------------------------------------------------------+
'debug +R7/R3 reply
                wrlong  reply,            ptrcmd        '\\                     save +R7/R3 reply
                add     ptrcmd,           #4            '//
:end            wrlong  xFFs, ptrcmd                    '\\                     FFFFFFFF
                add     ptrcmd, #4                      '//
'+-----------------------------------------------------------------------------+
getreply_ret
cmd_ret         ret
'+-----------------------------------------------------------------------------+

        
'+-----------------------------------------------------------------------------+
'+      SPI Routines...  (write & read together, 8/16/32 bits)                 +
'+-----------------------------------------------------------------------------+
sendio          rol     dataout,          #24           ' call here to Send a Byte (msbit first)
sendiohi        mov     bitsout,          #8            ' call here to Send 8x1's  / Receive a Byte
sendiocnt       mov     reply,            #0            ' call here to Send a Long / Receive a Long
:nextbit        andn    outa,             clkmask       ' CLK=0
                rol     dataout,          #1    wc      ' \ (write output bit)
                muxc    outa,             dimask        ' / set DI=0/1
                or      outa,             clkmask       ' CLK=1
                test    domask,           ina   wc      ' \ (read input bit)
                addx    reply,            reply         ' / accum DO input bits (Reply*2+C)
                djnz    bitsout,          #:nextbit     ' 8/16/32 bits?
sendio_ret
sendiohi_ret
sendiocnt_ret   ret

'+-----------------------------------------------------------------------------+
'+      Check Timeout...                                                       +
'+-----------------------------------------------------------------------------+
checktime       mov     duration,         cnt
                sub     duration,         starttime
                cmp     duration,         clockfreq wc
        if_nc   jmp     #fail_time                      ' failure:     timeout
checktime_ret   ret
        
'+-----------------------------------------------------------------------------+
'+      Failed...                                                              +
'+-----------------------------------------------------------------------------+
fail_time       add     x3333_0000,       #1
fail_readblock  add     x3333_0000,       #1
                add     x3333_0000,       #6
fail09          add     x3333_0000,       #1
fail08          add     x3333_0000,       #1
fail07          add     x3333_0000,       #1
fail06          add     x3333_0000,       #1
fail05          add     x3333_0000,       #1
fail04          add     x3333_0000,       #1
fail03          add     x3333_0000,       #1
fail02          add     x3333_0000,       #1
fail01          add     x3333_0000,       #1
fail00          wrlong  x3333_0000,       ptrcmd
                add     ptrcmd,           #4
                mov     SDtype,           #$F           ' ?? unknown card type
'+-----------------------------------------------------------------------------+
alldone         or      outa,             csmask        ' /CS=1
'+-----------------------------------------------------------------------------+
                wrlong  x99s,             ptrcmd        '\\                     99999999 = eof
                add     ptrcmd,           #4            '//
                wrlong  SDtype,           ptrcmd        '\\                     SD card type
                add     ptrcmd,           #4            '//
'+-----------------------------------------------------------------------------+
                wrlong  zero,             ptrdone       '                       set done
'+-----------------------------------------------------------------------------+
:halt           jmp     #:halt
'+-----------------------------------------------------------------------------+

initmask        long    1<<cs | 1<<di | 0<<clk  ' initial output
dirmask         long    1<<cs | 1<<di | 1<<clk  '(DO is SD output)
csmask          long    1<<cs                   ' /CS
dimask          long    1<<di                   ' mosi (fm P1)
clkmask         long    1<<clk                  ' CLK
domask          long    1<<do                   ' miso (to P1)

clockfreq       long    _freq                   ' xtal * pll
delay10us       long    _freq / 100_000         ' 10us
delay1ms        long    _freq / 1_000           ' 1ms

sectorsz        long    512
x000001AA       long    $0000_01AA              ' CMD8 Parameter
bit30                                           '\\
x40000000       long    $4000_0000              '// 1<<30 = bit30
xFFs            long    $FFFF_FFFF
xFFF            long    $FFF
xFF             long    $FF
x00                                             '\\
zero            long    0                       '//

x3333_0000      long    $3333_0000              ' debug fail nn
x99s            long    $9999_9999              '---eof---
x88s            long    $8888_8888              ' valid data token $FE (CMD17)
x77s            long    $7777_7777              ' decimal count follows...
xCS1            long    $FF11_11FF              ' /CS=1
SDtype          long    0                       ' 1=SDV1, 2=SDV2(byte address), 3=SDHC/SDV2(block address)
PTN0LBA         long    $0020_0000

cmdout          res     1                       ' The 8b CMDxx | $40
cmdpar          res     1                       ' The 32b parameters
cmdcrc          res     1                       ' The 8b CRC (must be valid for CMD0 & CMD8)
cmdpar2         res     1                       ' SDV1=$0, SDV2=$40000000
reply           res     1                       ' R1 reply (moved to reply8 when R3/R7 32b reply here)
reply8          res     1                       ' R1 reply (8b saved when R3/R7 32b reply follows)
dataout         res     1                       ' 8/32 bit data being shifted out
bitsout         res     1                       ' #bits to be shifted in/out
ctr1            res     1
ctr2            res     1

starttime       res     1
duration        res     1

bufad           res     1                       ' ptr sector buffer
blocknr         res     1                       ' sector#
blocksh         res     1                       ' block shift 0/9 bits

'following used for debugging
ptrdone         res     1 
ptrmbr          res     1 
ptrvol          res     1
ptrcmd          res     1
saveptrcmd      res     1

ctrbusy         res     1
ctr5541         res     1
time5541        res     1
timer           res     1